/*
MotorTask.c

Copyright 2001 Larry Barello

Authors: Larry Barello
        larry@barello.net
		Chris Baron
		chris@robot-power.com
*/
#include "io.h"
#include <progmem.h>
#include "sig-avr.h"
#include "avrx.h"
#include "hardware.h"
#include "MotorTask.h"
#include "LcdTask.h"

TimerControlBlock MotorTimer;

RCcontrolBlock RC0, RC1, RC2;

char    PWM_Mode,   // See eeprom.c for default behavior
        PWM_Rate,
        LeftSupply,
        RightSupply,
        MixedMode,
        DeadBand,
        SlewRate,
        BrakeMode,
		Flip_Mode,
		Flipped = 0;

char BrakeTimer;

DisableFlags fDisable;     // Global flags to kill output

/*+------------------------------------------------------------
char CheckPulse(pRCcontrolBlock)

Helper for calculating pulse width and validity
On exit the width contains a number from 0->0xFF

valid indicates if pulse is valid

Returns the inverse of "valid"
-*/

#define RCFAIL ((unsigned char)24)  // Number of loops until a pulse is considered failed.
#define MIN_RC_VALID    (8)    // Minimum # of valid pulses to see in period
#define MAX_RC_VALID    (16)    // Maximum # of valid pulses to see in period
#define MINVALIDWIDTH   (990/8) // Minimum valid pulse is 900uS long
#define MAXVALIDWIDTH   (2010/8)// Maximum valid pulse is 2000uS long

char CheckPulse(pRCcontrolBlock p, char CheckFail)
{
    if (CheckFail == 0)
    {
        if (MIN_RC_VALID < p->pulse && p->pulse < MAX_RC_VALID)   // assume 50hz update
        {
            p->valid = TRUE;
        }
        else
        {
            p->valid = FALSE;
        }
        p->pulse = 0;
    }
    
    if (p->done)
    {
        unsigned char tWidth;
        BeginCritical();
        p->rcpulse = tWidth = p->end - p->start;
        p->done = FALSE;            // Reset Pulse Trap
        EndCritical();

        // Exclude radio interference pulses < 1.0 ms long.
        
        if (MINVALIDWIDTH < tWidth && tWidth < MAXVALIDWIDTH)
//        if (tWidth >= MINVALIDWIDTH)
        {
            // First, deadband the center of the pulse
            // and shift the inputs appropriately
            
            unsigned char tCenter = (p->low + p->high + 1)/2;

            if (tWidth > tCenter + DeadBand)
                tWidth -= DeadBand;
            else if (tWidth < tCenter - DeadBand)
                tWidth += DeadBand;
            else
                tWidth = tCenter;

            if (tWidth > p->high)   // limit measured range.
                tWidth = p->high;
            else if (tWidth < p->low)
                tWidth = p->low;

            // Scale limited range to 0-0xFF (8 bit PWM) and store

            p->width = ((unsigned)(tWidth - p->low) * 0x0FF)/(unsigned)(p->high - p->low);

            p->pulse++;
        }
        else
        {
            p->pulse = 0;           // pulses
        }
    }
    return (!p->valid);   // Return True if not valid
}

/*+---------------------------------------------------------

int LimitDrive(int Drive)

Returns the passed parameter limited to the range of a signed 8 bit number

-*/
int LimitDrive(int Drive)
{
    if (Drive > 127)
        Drive = 127;
    if (Drive < -128)
        Drive = -128;
    return Drive;
}

// To cover math rounding errors and scaling errors (granularity, really)
// we kill the lower two bits of the output function

// changed to +- 3 5/11/02 to kill jitter around zero (LB)

int DoDeadband(int drive)
{
    if (-3 < drive && drive < 3)
    {
        drive = 0;
    }
    return (drive);
}
/*+--------------------------------------------------------------------------
    char DoSlew()

 Perform slew rate calculations to limit the rate of change in control signals.

 Requested drive is a signed char charactor.
 Slew rate is a posative integer from 0, 5, 10, 15, 20 and 25.
 Need to do integer math on the intermediate values.

 */

SlewType FLASH SlewTable[] = { { 0,127},
                               { 5, 26},
                               {10, 13},
                               {15,  8},
                               {20,  6},
                               {25,  5}
                             };
                           
char DoSlew(char RequestedDrive, pRCcontrolBlock p, char SlewRate)
{
    int CurrentDrive = p->Drive;
    
    SlewRate = __lpm_inline((unsigned)&SlewTable[(int)SlewRate].amount);

    if (RequestedDrive > CurrentDrive)
    {
        CurrentDrive += SlewRate;
        if (CurrentDrive > RequestedDrive)
        {
            CurrentDrive = RequestedDrive;
        }
    }
    else
    {
        CurrentDrive -= SlewRate;
        if (CurrentDrive < RequestedDrive)
        {
            CurrentDrive = RequestedDrive;
        }
    }
    p->Drive = CurrentDrive;
    return CurrentDrive;
}

/*+----------------------------------------------------
TASK!

MotorTask manages the R/C pulse measurements and converting them to the
appropriate PWM values for the hardware.

NB drive values are stored and manipulated as unsigned values in the
left/right RC structures, but handled as int's, below.

-*/

AVRX_GCC_TASKDEF(MotorTask, 16, 2)  // Priority == 2, run high priority.
{
    // NB: GCC/AvrX limitation, no more than 8 bytes of locals
    union
    {
        struct
        {
            char lower;
            char upper;
        };
        int Drive;
    }
    Left, Right;
    char PulseFailTimer = 0;
    unsigned char AuxLoHyst = 0;
    unsigned char AuxHiHyst = 0;
    struct
    {
        unsigned char Left:1;       // Implements variable brake
        unsigned char Right:1;
    }
    Coast;
    // NB: Make sure no more than 8 bytes of data declared, above!

    // Initialize the HIP chip here. All drive outputs are low, Disable is high.
    // so just drop disable for a couple hundred microseconds and then raise.

    AvrXDelay(&MotorTimer, MILLISECOND(500));// Delay 1/2 second before initializing.
    LEFT_ENABLE();
    RIGHT_ENABLE();
    AvrXDelay(&MotorTimer, 1);
    LEFT_DISABLE();
    RIGHT_DISABLE();

    while(1)
    {
        AvrXStartTimer(&MotorTimer, MILLISECOND(10));    // Poll @ 100hz

        fDisable.invalid_mode =
        fDisable.invalid_rate =
        fDisable.invalid_mix =
        fDisable.invalid_rc0 =
        fDisable.invalid_rc1 =
        fDisable.invalid_rc2 = FALSE;     // Reset kill bits
        
        switch(PWM_Rate)
        {
            case PWM2K:
                outp(TCCR1A_8BIT,  TCCR1A); // 8 bit normal clk/8
                outp(TCCR1B_CLK8,  TCCR1B);
                break;
            case PWM4K:
                outp(TCCR1A_8BIT,  TCCR1A); // 8 bit double rate clk/8
                outp(TCCR1B_CLK8D,  TCCR1B);
                break;
//            case PWM8K:
 //               outp(TCCR1A_9BIT,  TCCR1A); // Not supported
  //              outp(TCCR1B_CLK1,  TCCR1B);
   //             break;
            case PWM16K:
                outp(TCCR1A_8BIT,  TCCR1A); // 8 bit normal clk/1
                outp(TCCR1B_CLK1,  TCCR1B);
                break;
            case PWM32K:
                outp(TCCR1A_8BIT,  TCCR1A); // 8 bit double clk/1
                outp(TCCR1B_CLK1D,  TCCR1B);
                break;
            default:
                fDisable.invalid_rate = TRUE;
                break;
        }
        if (++PulseFailTimer > RCFAIL)      // Pass a flag to CheckPulse()
        {                                   // every 100ms
            PulseFailTimer = 0;
        }

        fDisable.invalid_rc0 = CheckPulse(&RC0, PulseFailTimer);
        fDisable.invalid_rc1 = CheckPulse(&RC1, PulseFailTimer);
        (void)CheckPulse(&RC2, PulseFailTimer);    // Aux Port

        // Convert input pulse to 8 bit signed value and mixed, if needed

        switch(MixedMode)
        {
            case MIXED:
                Left.Drive = LimitDrive(RC0.width+RC1.width-0x0100);
                Right.Drive = LimitDrive(RC0.width-RC1.width);
                break;
            case STRAIGHT:
                Left.Drive  = RC1.width - 0x080;
                Right.Drive = RC0.width - 0x080;
                break;
            default:
                fDisable.invalid_mix = TRUE;    // Disable output drive
                break;
        }

		// Now check for flipped condition and invert the drive
		// if flipped
		if (Flip_Mode && !FLIP_IN())
		{
			Flipped = 1;
			//Left.Drive = -Left.Drive;
			//Right.Drive = -Right.Drive;
            int temp = -Left.Drive;
			Left.Drive = -Right.Drive;
			Right.Drive = temp;
		}
		else {
			Flipped = 0;
		}

        // Force values near zero to zero (hide jitter)
        
        Left.Drive  = DoDeadband(Left.Drive);
        Right.Drive = DoDeadband(Right.Drive);
        
        // Implement Slew Rate Limit

        Left.Drive = DoSlew(Left.Drive, &RC1, SlewRate);
        Right.Drive = DoSlew(Right.Drive, &RC0, SlewRate);

        // Implement variable brake mode

        if (++BrakeTimer > 5)      // 1-10
            BrakeTimer = 1;

        Coast.Left = (Left.Drive == 0 && BrakeTimer > BrakeMode);
        Coast.Right = (Right.Drive == 0 && BrakeTimer > BrakeMode);

        // Adjust the output per the drive mode
        
        switch(PWM_Mode)
        {
            case LOCKED_ANTIPHASE:
                LOCKED_ANTI();
            
                Left.Drive += 0x080;     // Convert back to unsigned.
                Right.Drive += 0x080;
                break;
            case SIGN_MAGNITUDE:
                SIGN_MAG();
            
                if (Left.Drive >=0)
                {
                    LEFT_FORWARD();
                }
                else
                {
                    LEFT_REVERSE();
                }
                if (Right.Drive >=0)
                {
                    RIGHT_FORWARD();
                }
                else
                {
                    RIGHT_REVERSE();
                }
                Left.Drive  += Left.Drive;      // Double for PWM
                Right.Drive += Right.Drive;
                
                if (Left.Drive == 0x00FE)       // Ajust that last count..
                    Left.Drive = 0x00FF;
                if (Right.Drive == 0x00FE)
                    Right.Drive = 0x00FF;
                    
                break;
            default:
                fDisable.invalid_mode = TRUE;    // Disable output
                break;
        }
        
        if (Coast.Left || fDisable.Output)    // If any bits set, disable output
            LEFT_DISABLE();
        else
            LEFT_ENABLE();

        if (Coast.Right || fDisable.Output)
            RIGHT_DISABLE();
        else
            RIGHT_ENABLE();

        // For simplicity we deal with 8 bit PWM only
        
        outp(Left.lower, LEFT_DRIVEL);
        outp(Right.lower, RIGHT_DRIVEL);

//        RC0.Drive = Right.Drive;        // For debug purposes
//        RC1.Drive = Left.Drive;
        RC2.Drive = RC2.width;

        // Handle AUX outputs.
        
        if (RC2.valid)
        {
            if (RC2.width < AUXLOW_THRESHOLD + AuxLoHyst)
            {
                AUXLOW_ON();
                AuxLoHyst = 10;
            }
            else
            {
                AUXLOW_OFF();
                AuxLoHyst = 0;
            }
            if (RC2.width > AUXHIGH_THRESHOLD - AuxHiHyst)
            {
                AUXHIGH_ON();
                AuxHiHyst = 10;
            }
            else
            {
                AUXHIGH_OFF();
                AuxHiHyst = 0;
            }
        }
        else
        {
            AUXHIGH_OFF();
            AUXLOW_OFF();
            AuxHiHyst = AuxLoHyst = 0;
        }

        AvrXWaitTimer(&MotorTimer);
    }
}

